SysprofVisualizer {
  background: @content_view_bg;
  }
SysprofVisualizer:not(:last-child) {
  border-bottom: 1px solid alpha(@borders, 0.3);
  }

SysprofVisualizerGroup {
  border-bottom: 1px solid @borders;
  }
SysprofVisualizerGroup:last-child {
  box-shadow: 0 20px 15px 15px alpha(@borders, 0.3);
  }

SysprofVisualizersFrame box.horizontal.inline-toolbar {
  padding: 0;
  margin: 0;
  border: none;
  border-radius: 0;
  border-width: 0;
  }

SysprofVisualizersFrame viewport.visualizers {
  border-right: 1px solid @borders;
  box-shadow: 1px 1px 3px alpha(@borders, 0.5);
  }

SysprofVisualizersFrame scrollbar.horizontal {
  color: mix(@theme_fg_color, @theme_selected_bg_color, 0.5);
  background: transparent;
  }

SysprofVisualizersFrame scrollbar.horizontal range trough {
  background: transparent;
  }

SysprofVisualizersFrame scrollbar.horizontal range trough slider {
  margin-left: 1px;
  margin-right: 1px;
  padding: 6px;
  min-height: 14px;
  background: alpha(@content_view_bg, 0.2);
  border-radius: 3px;
  border: 2px solid @theme_selected_bg_color;
  box-shadow: inset 0 10px 5px alpha(@content_view_bg,.3),
              inset 0 -15px 5px alpha(@content_view_bg,.1),
              0px 2px 4px @borders;
  }

SysprofVisualizerTicks {
  color: mix(@theme_fg_color, @borders, 0.5);
  background-color: @content_view_bg;
  }

SysprofVisualizersFrame list {
  background-color: @theme_bg_color;
  }

SysprofVisualizersFrame list.visualizer-groups row {
  padding: 0;
  border-bottom: 1px solid @borders;
  }
SysprofVisualizersFrame list.visualizer-groups row:not(:selected) {
  background-color: @theme_bg_color;
  }
SysprofVisualizersFrame list.visualizer-groups row:last-child {
  box-shadow: 0 20px 15px 15px alpha(@borders, 0.3);
  }

SysprofVisualizersFrame .left-column .small-button.flat {
  border-color: transparent;
  min-height: 8px;
  min-width: 8px;
  }
SysprofVisualizersFrame .left-column .small-button.flat:checked,
SysprofVisualizersFrame .left-column .small-button.flat:hover {
  border-color: @borders;
  }

SysprofVisualizersFrame .selection {
  border-radius: 3px;
  background-color: alpha(@theme_selected_bg_color, 0.35);
  box-shadow: inset 0 10px 5px alpha(@content_view_bg,.3),
              inset 0 -15px 5px alpha(@content_view_bg,.1),
              inset 0 0 1px 1px @theme_selected_bg_color;
  }
