list.environ-editor row button.flat:last-child {
  min-height: 16px;
  min-width: 16px;
  padding: 0;
  margin: 3px;
}
list.environ-editor row button.flat:not(:hover) {
  border-color: transparent;
}
list.environ-editor row button.flat image {
  padding: 3px;
  margin: 0;
}
